#  Copyright © 2021 CloudBlue. All rights reserved.
import pathlib
import os
from datetime import datetime
from functools import partial

import pytz

from weasyprint import CSS, HTML, default_url_fetcher

from connect.reports.renderers.base import temp_dir
from connect.reports.renderers.j2 import Jinja2Renderer
from connect.reports.renderers.registry import register


def local_fetcher(url, root_dir=None, template_dir=None, cwd=None):
    tpl_dir_path = pathlib.Path(os.path.abspath(root_dir)) / pathlib.Path(template_dir)
    tpl_dir_url = tpl_dir_path.as_uri()
    if url.startswith(f'file://{cwd}'):
        rel_path = os.path.relpath(url[7:], cwd)
        new_path = os.path.join(
            os.path.abspath(root_dir),
            rel_path,
        )
        url = f'file://{new_path}'
    elif url.startswith(tpl_dir_url):
        count = url.count(template_dir)
        if url.count(template_dir) > 1:
            url = url.replace(f'{template_dir}/', '', count - 1)
    return default_url_fetcher(url)


@register('pdf')
class PDFRenderer(Jinja2Renderer):
    """
    PDF Renderer class.
    Inherits from BaseRenderer class and implements
    the generation report function, exporting the data
    to a PDF file.
    """
    def render(self, data, output_file, start_time=None):
        start_time = start_time or datetime.now(tz=pytz.utc)
        with temp_dir() as tmpdir:
            self.cwd = tmpdir
            report_file = self.generate_report(data, f'{tmpdir}/report')
            summary_file = self.generate_summary(f'{tmpdir}/summary', start_time)
            pack_file = self.pack_files(report_file, summary_file, output_file)
        return pack_file

    def generate_report(self, data, output_file):
        tokens = output_file.split('.')
        if tokens[-1] != 'pdf':
            output_file = f'{tokens[0]}.pdf'

        rendered_file = super().generate_report(data, output_file)
        fetcher = partial(
            local_fetcher,
            root_dir=self.root_dir,
            template_dir=os.path.dirname(self.template),
            cwd=getattr(self, 'cwd', os.getcwd()),
        )
        kwargs = {}
        css_file = self.args.get('css_file')
        if css_file:
            css = CSS(filename=os.path.join(self.root_dir, css_file), url_fetcher=fetcher)
            kwargs['stylesheets'] = [css]
        html = HTML(filename=rendered_file, url_fetcher=fetcher)
        html.write_pdf(output_file, **kwargs)
        return output_file

    @classmethod
    def validate(cls, definition):
        errors = super(PDFRenderer, cls).validate(definition)
        if definition.args is not None:
            css_file = definition.args.get('css_file')
            if css_file and not os.path.isfile(
                os.path.join(definition.root_path, css_file),
            ):
                errors.append(
                    f'css_file `{css_file}` not found.',
                )
        return errors



#  Copyright © 2021 CloudBlue. All rights reserved.

import os
from dataclasses import asdict, dataclass, field
from functools import cached_property
from typing import Any, Dict, List


@dataclass
class Account:
    id: str
    name: str


@dataclass
class Report:
    id: str
    name: str
    description: str
    values: List[Dict[str, Any]]


@dataclass
class RendererDefinition:
    """
    Renderer representation on `reports.json` file descriptor.

    :param root_path: Base root path.
    :type root_path: str
    :param id: Unique renderer identifier.
    :type id: str
    :param type: Renderer type.
    :type type: str
    :param description: Brief explanation about renderer.
    :type description: str
    :param default: Active/Inactive renderer.
    :type default: bool
    :param template: Template name
    :type template: str
    :param args: Additional arguments for using the renderer.
    :type args: dict
    """
    root_path: str
    id: str
    type: str
    description: str
    default: bool = False
    template: str = field(default=None)
    args: Dict[str, Any] = field(default=None)


@dataclass
class ParameterDefinition:
    """
    Parameter representation on `reports.json` file descriptor.

    :param id: Unique parameter identifier.
    :type id: str
    :param type: Parameter type.
    :type type: str
    :param name: Parameter name.
    :type name: str
    :param description: Brief explanation about parameter.
    :type description: str
    :param required: Required parameter.
    :type required: bool
    """
    id: str
    type: str
    name: str
    description: str
    required: bool = False


@dataclass
class ChoicesParameterDefinition(ParameterDefinition):
    choices: List[Dict[str, str]] = field(default_factory=list)


@dataclass
class ReportDefinition:
    """
    Report representation on `reports.json` file descriptor.

    :param root_path: Base root path.
    :type root_path: str
    :param name: Report name.
    :type name: str
    :param readme_file: Brief explanation about report.
    :type readme_file: str
    :param entrypoint: Function responsible to generate the report.
    :type entrypoint: str
    :param audience: Report audience.
    :type audience: list
    :param report_spec: Report specification.
    :type report_spec: str
    :param renderers: List of available renderers for report.
    :type renderers: list
    :param parameters: List of report parameters.
    :type parameters: list
    """
    root_path: str
    name: str
    readme_file: str
    entrypoint: str
    audience: List[str]
    report_spec: str
    renderers: List[RendererDefinition]
    parameters: List[ParameterDefinition] = field(default_factory=list)

    @cached_property
    def default_renderer(self):
        for renderer in self.renderers:
            if renderer.default:
                return renderer.id

    @property
    def local_id(self):
        tokens = self.entrypoint.split('.')
        return tokens[1] if len(tokens) > 2 else None

    @property
    def description(self):
        path = os.path.join(self.root_path, self.readme_file)
        with open(path, 'r') as fp:
            return fp.read()

    def get_parameters(self):
        return [
            asdict(p)
            for p in self.parameters
        ]

    def get_renderers(self):
        renderer_list = []
        for renderer in self.renderers:
            renderer_list.append(
                {
                    'id': renderer.id,
                    'type': renderer.type,
                    'description': renderer.description,
                    'default': renderer.default,
                    'template': renderer.template,
                    'args': renderer.args,
                },
            )
        return renderer_list


@dataclass
class RepositoryDefinition:
    """
    Report repository representation on `reports.json` file descriptor.

    :param root_path: Base root path.
    :type root_path: str
    :param readme_file: Brief explanation about report repository.
    :type readme_file: str
    :param name: Repository name.
    :type name: str
    :param version: Repository version.
    :type version: str
    :param language: Supported language for report generation.
    :type language: str
    :param reports: List of available reports.
    :type reports: list
    """
    root_path: str
    readme_file: str
    name: str
    version: str
    language: str = 'python'
    reports: List[ReportDefinition] = field(default_factory=list)

    @property
    def description(self):
        path = os.path.join(self.root_path, self.readme_file)
        with open(path, 'r') as fp:
            return fp.read()



#  Copyright © 2021 CloudBlue. All rights reserved.

from connect.reports.renderers.base import BaseRenderer
from connect.reports.renderers.registry import register

import csv


@register('csv')
class CSVRenderer(BaseRenderer):
    """
    CSV Renderer class.
    Inherits from BaseRenderer class and implements
    the generation report function, exporting the data
    to a CSV file.
    """
    def generate_report(self, data, output_file):
        tokens = output_file.split('.')
        if tokens[-1] != 'csv':
            output_file = f'{tokens[0]}.csv'
        with open(output_file, 'w') as fp:
            writer = csv.writer(fp, delimiter=';', quotechar='"', quoting=csv.QUOTE_ALL)
            for row in data:
                writer.writerow(row)
        return output_file



#  Copyright © 2021 CloudBlue. All rights reserved.

from connect.reports.renderers.registry import (  # noqa
    get_renderers,
    get_renderer_class,
    get_renderer,
)
from connect.reports.renderers.j2 import Jinja2Renderer # noqa
from connect.reports.renderers.json import JSONRenderer # noqa
from connect.reports.renderers.pdf import PDFRenderer # noqa
from connect.reports.renderers.xlsx import XLSXRenderer # noqa
from connect.reports.renderers.csv import CSVRenderer # noqa



#  Copyright © 2021 CloudBlue. All rights reserved.
import contextlib
import json
import os
import shutil
import tempfile
import zipfile
from abc import ABCMeta, abstractmethod
from datetime import datetime

import pytz


@contextlib.contextmanager
def temp_dir():
    name = tempfile.mkdtemp()
    yield name
    try:
        if os.path.isdir(name):
            shutil.rmtree(name)
    except Exception:
        pass


class BaseRenderer(metaclass=ABCMeta):
    """
    Base renderer class with minimum required functionality
    that renderers have to inherit and implement.

    :param environment: Runtime environment.
    :type environment: str
    :param root_dir: Base root dir.
    :type root_dir: str
    :param account: Owner account.
    :type account: Account
    :param report: Report object.
    :type report: Report
    :param template: Template name.
    :type template: str
    :param args: Renderer required arguments.
    :type args: dict
    :param extra_context: Additional context data
                        for report rendering.
    :type extra_context: dict
    """
    def __init__(
        self,
        environment,
        root_dir,
        account,
        report,
        template=None,
        args=None,
    ):
        self.environment = environment
        self.root_dir = root_dir
        self.account = account
        self.report = report
        self.template = template
        self.args = args or {}
        self.extra_context = None

    def get_context(self, data):
        context = {
            'account': self.account,
            'report': self.report,
            'data': data,
        }
        if self.extra_context:
            context['extra_context'] = self.extra_context
        return context

    def set_extra_context(self, data):
        self.extra_context = data

    def render(self, data, output_file, start_time=None):
        """
        Creates effectively report pack file (report + summary files)

        :param data: Report information.
        :type data: dict
        :param output_file: Output file name.
        :type output_file: str
        :param start_time: Start time information.
        :type start_time: datetime
        """
        start_time = start_time or datetime.now(tz=pytz.utc)
        with temp_dir() as tmpdir:
            report_file = self.generate_report(data, f'{tmpdir}/report')
            summary_file = self.generate_summary(f'{tmpdir}/summary', start_time)
            pack_file = self.pack_files(report_file, summary_file, output_file)
        return pack_file

    def generate_summary(self, output_file, start_time):
        """
        Generates summary information of report generation.

        :param output_file: Output file name.
        :type output_file: str
        :param start_time: Start time information.
        :type start_time: datetime
        """
        data = {
            'title': 'Report Execution Information',
            'data': {
                'report_start_time': start_time.isoformat(),
                'report_finish_time': datetime.now(tz=pytz.utc).isoformat(),
                'account_id': self.account.id,
                'account_name': self.account.name,
                'report_id': self.report.id,
                'report_name': self.report.name,
                'runtime_environment': self.environment,
                'report_execution_parameters': self.report.values,
            },
        }
        output_file = f'{output_file}.json'
        json.dump(data, open(output_file, 'w'), indent=4, sort_keys=True)
        return output_file

    def pack_files(self, report_file, summary_file, output_file):
        tokens = output_file.split('.')
        if tokens[-1] != 'zip':
            output_file = f'{tokens[0]}.zip'
        with zipfile.ZipFile(output_file, 'w', compression=zipfile.ZIP_DEFLATED) as repzip:
            repzip.write(report_file, os.path.basename(report_file))
            repzip.write(summary_file, os.path.basename(summary_file))

        return output_file

    @abstractmethod
    def generate_report(self, data, output_file):
        """
        Method to be implemented by the specific renderer.
        Generates report file.

        :param data: Report information.
        :type data: dict
        :param output_file: Output file name.
        :type output_file: str
        """
        raise NotImplementedError('Subclasses must implement the `generate_report` method.')

    @classmethod
    def validate(cls, definition):
        return []



#  Copyright © 2021 CloudBlue. All rights reserved.

import json
import os
from collections import Counter

import jsonschema

from connect.reports.renderers import get_renderer_class, get_renderers


CHOICES_PARAM_TYPE = [
    'checkbox',
    'choice',
]

ALL_PARAM_TYPES = [
    'date_range',
    'date',
    'single_line',
    'object',
    'product',
    'marketplace',
    'hub',
    *CHOICES_PARAM_TYPE,
]

JSON_REPORTS_SCHEMA = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'schemas/reports_schema.json',
)


def _get_duplicates(data):
    return [
        item for item, count in Counter(data).items() if count > 1
    ]


def validate_with_schema(json_data):
    """
    Validates reports descriptor content against json schema.

    :param json_data: Reports descriptor string content.
    :type json_data: dict
    :returns: A string list of errors if exist.
    :rtype: str
    """
    with open(JSON_REPORTS_SCHEMA, 'r') as fp:
        json_schema = json.load(fp)
    try:
        jsonschema.validate(instance=json_data, schema=json_schema)
    except jsonschema.exceptions.ValidationError as errors:
        return str(errors)


def _validate_parameters(report_id, parameters):
    """
    Validates report parameters.

    :param report_id: Report unique identifier.
    :type report_id: str
    :param parameters: List of report parameters.
    :type parameters: list
    :returns: A list of errors.
    :rtype: list
    """
    errors = []
    params_ids = []
    for param in parameters:
        params_ids.append(param.id)

        if param.type not in ALL_PARAM_TYPES:
            errors.append(
                f'Invalid type for parameter `{param.id}` on report `{report_id}`: `{param.type}`.',
            )

        if param.type in CHOICES_PARAM_TYPE and not param.choices:
            errors.append(f'Missing choices for parameter `{param.id}` on report `{report_id}`.')

    diff = set(_get_duplicates(params_ids))
    if diff:
        errors.append(
            f'The following parameter ids are duplicated on report `{report_id}`: {",".join(diff)}',
        )
    return errors


def _validate_renderer(report_id, renderer):
    """
    Validates renderer.

    :param report_id: Report unique identifier.
    :type report_id: str
    :param renderer: Renderer definition object.
    :type renderer: RendererDefinition
    :returns: A list of errors.
    :rtype: list
    """
    errors = []
    if renderer.type not in get_renderers():
        errors.append(
            f'renderer `{renderer.id}` of type `{renderer.type}` is not known on `{report_id}`.',
        )
        return errors

    renderer_cls = get_renderer_class(renderer.type)
    errors.extend(renderer_cls.validate(renderer))
    return errors


def _validate_report(report):
    """
    Validates report.

    :param report: Report definition object.
    :type report: ReportDefinition
    :returns: A list of errors.
    :rtype: list
    """
    errors = []
    if not os.path.isfile(
        os.path.join(report.root_path, report.readme_file),
    ):
        errors.append(
            f'report `{report.local_id}` property `readme_file` cannot be resolved to a file.',
        )

    tokens = report.entrypoint.split('.')
    if len(tokens) < 2:
        errors.append(
            f'entrypoint `{report.entrypoint}` does not follow the package structure.',
        )
        return errors

    report_root = os.path.join(report.root_path, tokens[0], tokens[1])
    if not (os.path.isdir(report_root) or os.path.isfile(f'{report_root}.py')):
        errors.append(
            f'entrypoint `{report.entrypoint}` directory structure '
            'does not match the package definition.',
        )

    errors.extend(_validate_parameters(report.local_id, report.parameters))

    renderers_ids = []
    default_renderers = []
    for renderer in report.renderers:
        if renderer.default:
            default_renderers.append(renderer.id)
        renderers_ids.append(renderer.id)
        errors.extend(_validate_renderer(report.local_id, renderer))

    diff = set(_get_duplicates(renderers_ids))
    if diff:
        errors.append(f'report {report.local_id} has duplicated renderer ids: {",".join(diff)}')

    if len(default_renderers) > 1:
        errors.append(
            f'report {report.local_id} has multiple default renderers: '
            f'{",".join(default_renderers)}',
        )

    return errors


def validate(repo):
    """
    Validates whole repository definition.

    :param repo: Repository definition object.
    :type repo: RepositoryDefinition
    :returns: A list of errors.
    :rtype: list
    """
    errors = []
    if not os.path.isfile(
        os.path.join(repo.root_path, repo.readme_file),
    ):
        errors.append(
            'repository property `readme_file` cannot be resolved to a file.',
        )

    reports_local_ids = []
    for report in repo.reports:
        reports_local_ids.append(report.local_id)
        errors.extend(_validate_report(report))

    diff = set(_get_duplicates(reports_local_ids))
    if diff:
        errors.append(
            f'Multiple reports within single module found: {",".join(diff)}',
        )

    return errors
